"""
Copyright (C) 2026 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# pylint: disable=arguments-differ

from typing import List

import numpy as np
from onnx.helper import make_node, tensor_dtype_to_np_dtype
from onnx.onnx_pb import NodeProto

from ... import OnnxGraph
from .. import PASSES
from ..pattern import SingleNodePattern
from ..rewriter import Rewriter
from ..utils import make_constant


@PASSES.register(name="space2depth_to_conv")
class SpaceToDepthToConvRewriter(Rewriter):
    """Convert SpaceToDepth to Conv."""

    def __init__(self):
        super().__init__(pattern=SingleNodePattern("SpaceToDepth"))

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto], to_depthwise=False):
        s2d = nodes[0]
        blocksize = self.get_attribute(s2d, "blocksize")
        assert isinstance(blocksize, int)
        _, ic, _, _ = graph.tensor_shape(s2d.input[0])
        _, oc, _, _ = graph.tensor_shape(s2d.output[0])
        assert isinstance(ic, int) and isinstance(oc, int)
        dtype = graph.tensor_type(s2d.input[0])
        assert ic * blocksize**2 == oc, "invalid space2depth parameters"

        assert not to_depthwise, "not implemented yet"
        if to_depthwise:
            kernel = np.tile(
                np.eye(blocksize**2)
                .reshape([-1, blocksize, blocksize])
                .astype(tensor_dtype_to_np_dtype(dtype)),
                [ic, 1, 1],
            )[:, None]
            # TODO: output channels need to be reordered
        else:
            kernel = (
                np.tile(
                    np.eye(blocksize**2)
                    .reshape([-1, blocksize, blocksize])
                    .astype(tensor_dtype_to_np_dtype(dtype)),
                    [ic, 1, 1, 1],
                )
                .transpose([1, 0, 2, 3])
                .reshape([oc, 1, blocksize, blocksize])
            )
            # expand to [oc, ic, blocksize, blocksize]
            kernel = np.concatenate(
                [kernel, np.tile(np.zeros_like(kernel), [1, ic - 1, 1, 1])], axis=1
            )
            for i in range(1, ic):
                kernel[i::ic][:, i] = kernel[i::ic][:, 0]
                kernel[i::ic][:, 0] = 0  # swap
        conv_weight = make_constant(name=f"{s2d.name}/weight", value=kernel)
        conv = make_node(
            op_type="Conv",
            inputs=list(s2d.input) + [conv_weight.output[0]],
            outputs=s2d.output,
            name=f"{s2d.name}/conv",
            kernel_shape=[blocksize, blocksize],
            strides=[blocksize, blocksize],
            group=ic if to_depthwise else 1,
        )
        self -= s2d
        self += [conv, conv_weight]



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# pylint: disable=missing-function-docstring

import numpy as np
import onnx
import pytest
from onnx.helper import (
    make_graph,
    make_model,
    make_node,
    make_operatorsetid,
    make_tensor_value_info,
)
from onnx.numpy_helper import from_array

from onnxifier.evaluator import Evaluator
from onnxifier.graph import OnnxGraph
from onnxifier.passes.version_converter.downgrade import downgrade_op_version


def _make_col2im_op18(
    x, image_shape, block_shape, dilations=None, pads=None, strides=None
):
    assert x.ndim == 3
    node = make_node(
        "Col2Im",
        ["input", "image_shape", "block_shape"],
        ["output"],
        "col2im",
        dilations=dilations,
        pads=pads,
        strides=strides,
    )
    graph = make_graph(
        [node],
        "graph",
        [
            make_tensor_value_info("input", onnx.TensorProto.FLOAT, list(x.shape)),
        ],
        [
            make_tensor_value_info(
                "output", onnx.TensorProto.FLOAT, ["B", "C", "H", "W"]
            )
        ],
        [
            from_array(np.array(image_shape, np.int64), "image_shape"),
            from_array(np.array(block_shape, np.int64), "block_shape"),
        ],
    )
    model = make_model(
        graph,
        ir_version=9,
        opset_imports=[make_operatorsetid("", 18)],
    )
    onnx.checker.check_model(model, True)
    return OnnxGraph(model)


def _numpy_impl(
    x: np.ndarray, image_shape, block_shape, dilations=None, pads=None, strides=None
):
    """Implement Col2Im using numpy with proper handling of padding and overlaps."""
    B, C, L = x.shape
    blocks = np.prod(block_shape)
    x = x.reshape([B, C // blocks, *block_shape, L])
    if dilations is None:
        dilations = np.ones_like(image_shape)
    if pads is None:
        pads = np.repeat(np.zeros_like(image_shape), 2)
    if strides is None:
        strides = np.ones_like(image_shape)
    vpad = np.array(pads, np.int64).reshape([2, -1]).T
    spatial_shape = []
    for ch, d, s, p, b in zip(image_shape, dilations, strides, vpad, block_shape):
        spatial_shape.append(
            np.ceil((ch + p[0] + p[1] - (d * (b - 1))) / s).astype(np.int64)
        )
    assert L == np.prod(spatial_shape)

    # canvas based filling
    y = np.zeros([B, C // blocks, *image_shape], dtype=x.dtype)
    for ch in range(y.shape[1]):
        # === Below this block-wise loop is generated by AI (Trae)
        block = x[:, ch].reshape([B, *block_shape, *spatial_shape])
        # Generate all combinations of kernel indices
        kernel_indices = np.meshgrid(
            *[np.arange(d) for d in block_shape], indexing="ij"
        )
        kernel_indices = np.stack(kernel_indices, axis=-1).reshape(-1, len(block_shape))

        # Generate all combinations of spatial indices
        spatial_indices = np.meshgrid(
            *[np.arange(d) for d in spatial_shape], indexing="ij"
        )
        spatial_indices = np.stack(spatial_indices, axis=-1).reshape(
            -1, len(spatial_shape)
        )

        # Create all combinations of kernel and spatial indices
        kdx = np.repeat(kernel_indices, len(spatial_indices), axis=0)
        sdx = np.tile(spatial_indices, (len(kernel_indices), 1))

        # Calculate output positions using correct Col2Im formula
        output_positions = []
        for i in range(len(spatial_shape)):
            pos = sdx[:, i] * strides[i] + kdx[:, i] * dilations[i] - vpad[i][0]
            output_positions.append(pos)
        odx = np.stack(output_positions, axis=1)

        # Filter valid positions within output bounds
        output_shape = np.array(y.shape[2:])
        valid_mask = np.logical_and.reduce(
            [odx[:, i] >= 0 for i in range(len(spatial_shape))]
        ) & np.logical_and.reduce(
            [odx[:, i] < output_shape[i] for i in range(len(spatial_shape))]
        )

        # Apply valid mask
        valid_kernel = kdx[valid_mask]
        valid_spatial = sdx[valid_mask]
        valid_positions = odx[valid_mask]

        # Accumulate block values to output image
        if valid_positions.size > 0:
            # Create multi-dimensional indices for block
            block_idx = tuple(valid_kernel.T) + tuple(valid_spatial.T)
            # Create indices for output image y
            y_idx = (slice(None), ch) + tuple(valid_positions.T)
            # Accumulate values using numpy's scatter add
            np.add.at(y, y_idx, block[(slice(None),) + block_idx])
        # === Generation done

    return y


@pytest.fixture(
    params=[
        dict(
            shape=(1, 32, 625),
            image_shape=(24, 24),
            block_size=(2, 2),
            paddings=(1, 1, 1, 1),
            strides=(1, 1),
            output_shape=(1, 8, 24, 24),
        ),
        dict(
            shape=(1, 18, 144),
            image_shape=(12, 12),
            block_size=(3, 3),
            paddings=(1, 1, 1, 1),
            strides=(1, 1),
            output_shape=(1, 2, 12, 12),
        ),
        dict(
            shape=(1, 32, 144),
            image_shape=(24, 24),
            block_size=(2, 2),
            paddings=None,
            strides=(2, 2),
            output_shape=(1, 8, 24, 24),
        ),
    ]
)
def test_params(request):
    return request.param


def test_col2im_numpy_impl(test_params):
    shape = test_params["shape"]
    image_shape = test_params["image_shape"]
    block_size = test_params["block_size"]
    paddings = test_params["paddings"]
    strides = test_params["strides"]
    output_shape = test_params["output_shape"]

    x = np.random.randn(*shape).astype("float32")
    graph = _make_col2im_op18(
        x,
        image_shape=image_shape,
        block_shape=block_size,
        pads=paddings,
        strides=strides,
    )
    runner1 = Evaluator(graph.model)
    [y_ref] = runner1(["output"], input=x)

    # Compute our numpy implementation output
    y = _numpy_impl(
        x,
        image_shape=image_shape,
        block_shape=block_size,
        pads=paddings,
        strides=strides,
    )

    # Validate output shape and numerical accuracy
    assert y_ref.shape == output_shape
    # Temporarily relax tolerance for debugging
    assert np.allclose(
        y, y_ref, atol=1e-5
    ), f"Numerical mismatch: max diff {np.max(np.abs(y - y_ref)):.6f}"


def test_downgrade_col2im_op18(test_params):
    shape = test_params["shape"]
    image_shape = test_params["image_shape"]
    block_size = test_params["block_size"]
    paddings = test_params["paddings"]
    strides = test_params["strides"]
    output_shape = test_params["output_shape"]

    x = np.random.randn(*shape).astype("float32")
    graph = _make_col2im_op18(
        x,
        image_shape=image_shape,
        block_shape=block_size,
        pads=paddings,
        strides=strides,
    )
    runner1 = Evaluator(graph.model)
    [y_ref] = runner1(["output"], input=x)

    graph = downgrade_op_version(graph, op_version=17)
    runner2 = Evaluator(graph.model)
    [y] = runner2(["output"], input=x)

    assert y_ref.shape == output_shape
    assert y.shape == y_ref.shape
    assert np.allclose(y, y_ref)


def test_downgrade_col2im_to_depthtospace_op18():
    x = np.random.randn(1, 32, 144).astype("float32")
    graph = _make_col2im_op18(
        x, image_shape=[24, 24], block_shape=[2, 2], strides=[2, 2]
    )
    runner1 = Evaluator(graph.model)
    [y_ref] = runner1(["output"], input=x)

    graph = downgrade_op_version(graph, op_version=17)
    runner2 = Evaluator(graph.model)
    [y] = runner2(["output"], input=x)

    assert y_ref.shape == (1, 8, 24, 24)
    assert y.shape == (1, 8, 24, 24)
    assert np.allclose(y, y_ref)



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import warnings
from pathlib import Path

import click
import networkx as nx
import numpy as np
import onnx

from onnxifier import OnnxGraph, PassManager

# from onnx.helper import get_node_attr_value


def conv_analysis(
    graph: OnnxGraph,
    node: onnx.NodeProto,
):
    """Analyze Conv node.

    Args:
        graph (OnnxGraph): Graph.
        node (onnx.NodeProto): Node.

    Returns:
        dict: Analysis result.
    """
    ishape = graph.static_tensor_shape(node.input[0])
    wshape = graph.static_tensor_shape(node.input[1])
    oshape = graph.static_tensor_shape(node.output[0])
    if len(node.input) > 2 and node.input[2] != "":
        bias = graph.static_tensor_shape(node.input[2])
    else:
        bias = None
    # strides = get_node_attr_value(node, "strides")

    ic = ishape[1]
    oc = wshape[1]
    kh = wshape[2]
    kw = wshape[3]
    oh = oshape[2]
    ow = oshape[3]
    return dict(
        layer=node.name,
        op_type=node.op_type,
        ic=ic,
        oc=oc,
        kh=kh,
        kw=kw,
        idepth=1,
        ih=ishape[2],
        iw=ishape[3],
        oh=oh,
        ow=ow,
        act=0,
        order=0,
        ops=oc * oh * ow * (ic * kh * kw + (1 if bias else 0)),
        load=(
            np.prod(ishape) + np.prod(wshape) + (np.prod(bias) if bias else 0)
        ).item(),
        store=np.prod(oshape).item(),
    )


def matmul_analysis(
    graph: OnnxGraph,
    node: onnx.NodeProto,
):
    """Analyze MatMul node.

    Args:
        graph (OnnxGraph): Graph.
        node (onnx.NodeProto): Node.

    Returns:
        dict: Analysis result.
    """
    shape_a = graph.static_tensor_shape(node.input[0])
    shape_b = graph.static_tensor_shape(node.input[1])

    shape_a = [np.prod(shape_a[:-1]), shape_a[-1]]
    shape_b = [shape_b[0], np.prod(shape_b[1:])]
    m, n, k = shape_a[0], shape_b[1], shape_a[1]

    return dict(
        layer=node.name,
        op_type=node.op_type,
        ic=m,
        oc=n,
        kh=1,
        kw=1,
        idepth=1,
        ih=1,
        iw=k,
        oh=1,
        ow=k,
        act=0,
        order=0,
        ops=m * n * k,
        load=(np.prod(shape_a) + np.prod(shape_b)).item(),
        store=(m * n).item(),
    )


def unary_analysis(
    graph: OnnxGraph,
    node: onnx.NodeProto,
    no_ops: bool = False,
):
    """Analyze unary (activation) node.

    Args:
        graph (OnnxGraph): Graph.
        node (onnx.NodeProto): Node.

    Returns:
        dict: Analysis result.
    """
    ishape = graph.static_tensor_shape(node.input[0])
    oshape = graph.static_tensor_shape(node.output[0])
    while len(ishape) < 4:
        ishape.append(1)
    while len(oshape) < 4:
        oshape.append(1)
    return dict(
        layer=node.name,
        op_type=node.op_type,
        ic=ishape[1],
        oc=ishape[1],
        kh=1,
        kw=1,
        idepth=1,
        ih=ishape[2],
        iw=ishape[3],
        oh=oshape[2],
        ow=oshape[3],
        act=1,
        order=0,
        ops=0 if no_ops else np.prod(ishape).item(),
        load=np.prod(ishape).item(),
        store=np.prod(oshape).item(),
    )


def binary_analysis(
    graph: OnnxGraph,
    node: onnx.NodeProto,
):
    """Analyze Binary node.

    Args:
        graph (OnnxGraph): Graph.
        node (onnx.NodeProto): Node.

    Returns:
        dict: Analysis result.
    """
    shape = graph.static_tensor_shape(node.output[0])
    n_inputs = len(node.input)
    while len(shape) < 4:
        shape.append(1)

    return dict(
        layer=node.name,
        op_type=node.op_type,
        ic=shape[1],
        oc=shape[1],
        kh=1,
        kw=1,
        idepth=1,
        ih=shape[2],
        iw=shape[3],
        oh=shape[2],
        ow=shape[3],
        act=0,
        order=1,
        ops=np.prod(shape).item(),
        load=np.prod(shape).item() * n_inputs,
        store=np.prod(shape).item(),
    )


def resize_analysis(
    graph: OnnxGraph,
    node: onnx.NodeProto,
):
    """Analyze Resize node.

    Args:
        graph (OnnxGraph): Graph.
        node (onnx.NodeProto): Node.

    Returns:
        dict: Analysis result.
    """
    ishape = graph.static_tensor_shape(node.input[0])
    oshape = graph.static_tensor_shape(node.output[0])

    ic = ishape[1]
    oc = oshape[1]
    ih = ishape[2]
    iw = ishape[3]
    oh = oshape[2]
    ow = oshape[3]
    return dict(
        layer=node.name,
        op_type=node.op_type,
        ic=ic,
        oc=oc,
        kh=oh // ih,
        kw=ow // iw,
        idepth=1,
        ih=ih,
        iw=iw,
        oh=oh,
        ow=ow,
        act=0,
        order=1,
        ops=np.prod(oshape).item() * 4,
        load=np.prod(ishape).item(),
        store=np.prod(oshape).item(),
    )


def pool_analysis(
    graph: OnnxGraph,
    node: onnx.NodeProto,
):
    """Analyze Pool node.

    Args:
        graph (OnnxGraph): Graph.
        node (onnx.NodeProto): Node.

    Returns:
        dict: Analysis result.
    """

    ishape = graph.static_tensor_shape(node.input[0])
    oshape = graph.static_tensor_shape(node.output[0])

    ic = ishape[1]
    oc = oshape[1]
    ih = ishape[2]
    iw = ishape[3]
    oh = oshape[2]
    ow = oshape[3]
    return dict(
        layer=node.name,
        op_type=node.op_type,
        ic=ic,
        oc=oc,
        kh=oh // ih,
        kw=ow // iw,
        idepth=1,
        ih=ih,
        iw=iw,
        oh=oh,
        ow=ow,
        act=0,
        order=1,
        ops=np.prod(oshape).item(),
        load=np.prod(ishape).item(),
        store=np.prod(oshape).item(),
    )


def gridsample_analysis(
    graph: OnnxGraph,
    node: onnx.NodeProto,
):
    """Analyze GridSample node.

    Args:
        graph (OnnxGraph): Graph.
        node (onnx.NodeProto): Node.

    Returns:
        dict: Analysis result.
    """

    ishape = graph.static_tensor_shape(node.input[0])
    oshape = graph.static_tensor_shape(node.output[0])
    grid_shape = graph.static_tensor_shape(node.input[1])
    while len(ishape) < 5:
        ishape.append(1)
    while len(oshape) < 5:
        oshape.append(1)
    return dict(
        layer=node.name,
        op_type=node.op_type,
        ic=ishape[1],
        oc=oshape[1],
        kh=1,
        kw=1,
        idepth=ishape[-1],
        ih=ishape[2],
        iw=ishape[3],
        oh=oshape[2],
        ow=oshape[3],
        act=0,
        order=1,
        ops=4 * np.prod(grid_shape).item() * ishape[1],
        load=4 * np.prod(grid_shape).item() * ishape[1],
        store=np.prod(oshape).item(),
    )


def scatternd_analysis(
    graph: OnnxGraph,
    node: onnx.NodeProto,
):
    """Analyze ScatterND node.

    Args:
        graph (OnnxGraph): Graph.
        node (onnx.NodeProto): Node.

    Returns:
        dict: Analysis result.
    """

    update_shape = graph.static_tensor_shape(node.input[2])
    return dict(
        layer=node.name,
        op_type=node.op_type,
        ic=1,
        oc=1,
        kh=1,
        kw=1,
        idepth=1,
        ih=1,
        iw=1,
        oh=1,
        ow=1,
        act=0,
        order=1,
        ops=0,
        load=np.prod(update_shape).item(),
        store=np.prod(update_shape).item(),
    )


def gather_analysis(
    graph: OnnxGraph,
    node: onnx.NodeProto,
):
    """Analyze Gather node.

    Args:
        graph (OnnxGraph): Graph.
        node (onnx.NodeProto): Node.

    Returns:
        dict: Analysis result.
    """
    oshape = graph.static_tensor_shape(node.output[0])
    return dict(
        layer=node.name,
        op_type=node.op_type,
        ic=1,
        oc=1,
        kh=1,
        kw=1,
        idepth=1,
        ih=1,
        iw=1,
        oh=1,
        ow=1,
        act=0,
        order=1,
        ops=0,
        load=np.prod(oshape).item(),
        store=np.prod(oshape).item(),
    )


@click.command()
@click.option("-o", "--output", type=click.Path(dir_okay=False), required=True)
@click.argument(
    "model",
    type=click.Path(exists=True, dir_okay=False),
)
def analysis(model: None | Path = None, output: None | Path = None):
    assert model is not None
    graph = OnnxGraph(onnx.load_model(model))
    graph = PassManager(["infer_shape"]).optimize(graph)

    data = []
    for node_name in nx.topological_sort(graph):
        node: onnx.NodeProto = graph.nodes[node_name]["pb"]
        if node.op_type in ("Conv", "ConvTranspose"):
            data.append(conv_analysis(graph, node))
        elif node.op_type in ("MatMul",):
            data.append(matmul_analysis(graph, node))
        elif node.op_type in (
            "Relu",
            "Tanh",
            "Sigmoid",
            "HardSwish",
            "HardSigmoid",
            "Erf",
            "Transpose",
            "Sqrt",
            "Pow",
            "Clip",
        ):
            data.append(unary_analysis(graph, node))
        elif node.op_type in ("Tile",):
            data.append(unary_analysis(graph, node, no_ops=True))
        elif node.op_type in ("Add", "Sub", "Mul", "Div"):
            data.append(binary_analysis(graph, node))
        elif node.op_type in ("Resize",):
            data.append(resize_analysis(graph, node))
        elif node.op_type in ("MaxPool", "AveragePool"):
            data.append(pool_analysis(graph, node))
        elif node.op_type in ("GridSample",):
            data.append(gridsample_analysis(graph, node))
        elif node.op_type in ("ScatterND",):
            data.append(scatternd_analysis(graph, node))
        elif node.op_type in ("Gather", "GatherND", "GatherElements"):
            data.append(gather_analysis(graph, node))
        else:
            warnings.warn(f"Not support {node.op_type} now.")
    assert output is not None
    with open(output, "w", encoding="utf-8") as f:
        f.write(",".join(data[0].keys()) + "\n")
        for row in data:
            f.write(",".join(map(str, row.values())) + "\n")


if __name__ == "__main__":
    analysis()



"""
Copyright (C) 2024 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
from typing import Optional

import onnx

from .... import ONNXIFIER_IR_VERSION, ONNXIFIER_OPSET, PassManager
from ....graph import OnnxGraph
from .ir.ir11 import ir_to_onnx
from .passes.op import IR_PASSES, OP_CONVERT

_POST_PASSES = [
    "conv_dequantize_weight",
    "gemm_dequantize_weight",
    "recalculate_dequantize_weight",
    "eliminate_dead_nodes",
]


def openvino_xml_to_onnx_graph(
    model_path: str | os.PathLike | onnx.ModelProto,
    model_bin: Optional[str | os.PathLike] = None,
) -> onnx.ModelProto:
    """Convert OpenVINO IR .xml file to an equivalent ONNX graph.

    Args:
        model_path (str | Path): Path to the OpenVINO IR.xml file.
        model_bin (str | Path, optional): Path to the OpenVINO IR.bin file.
            Defaults to None. If model_bin is None, the binary is to be found in the
            same directory as the xml file with the same name but with a.bin extension.
    """
    if isinstance(model_path, onnx.ModelProto) and model_bin is None:
        ov_onnx = model_path  # for nested call but do not expose this interface
    elif not isinstance(model_path, onnx.ModelProto):
        ov_onnx = ir_to_onnx(model_path, model_bin)
    else:
        raise ValueError("model_bin is specified but model_path is ModelProto.")
    graph = OnnxGraph(ov_onnx)
    # sort OP_CONVERT here to make sure to put "Trivial" to the end
    passes = set(IR_PASSES).difference(OP_CONVERT)
    pm = PassManager(list(passes) + sorted(OP_CONVERT) + _POST_PASSES)
    model = pm.optimize(graph, strict=True).model
    model.ir_version = ONNXIFIER_IR_VERSION
    for opset in list(model.opset_import):
        if opset.domain not in ("", "ai.onnx", "ai.onnx.ml"):
            model.opset_import.remove(opset)
    if not model.opset_import:
        model.opset_import.append(ONNXIFIER_OPSET)
    return model



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# pylint: disable=arguments-differ

from typing import List

from onnx.helper import make_node
from onnx.onnx_pb import NodeProto

from ... import OnnxGraph, logger
from .. import PASSES
from ..pattern import SingleNodePattern
from ..rewriter import Rewriter


@PASSES.register("transpose_softmax_to_last_dim")
class TransposeSoftmaxToLastDimRewriter(Rewriter):
    """XNNC doesn't support Softmax with dim other than -1, we need to transpose the
    input tensor and output tensor to change the dim of Softmax to -1.

    Before:

        a->softmax(dim=1)->b

    After:

        a->transpose()->softmax(dim=-1)->transpose()->b
    """

    def __init__(self):
        super().__init__(SingleNodePattern("Softmax"))

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto], *args, **kwargs):
        node = nodes[0]
        axis = self.get_attribute(node, "axis")
        assert axis is None or isinstance(axis, (int, float, str))
        input_rank = len(graph.tensor_shape(node.input[0]))

        if axis is None or int(axis) in (-1, input_rank - 1):
            return

        axis = int(axis)
        if axis < 0:
            axis += input_rank
        perm = list(range(input_rank))
        perm[-1], perm[axis] = perm[axis], perm[-1]
        logger.debug(f"insert transpose before and after softmax: perm={perm}")

        trans_beg = make_node(
            "Transpose",
            inputs=[node.input[0]],
            outputs=[f"{node.name}/Transpose/beg_output0"],
            name=f"{node.name}/Transpose/beg",
            perm=perm,
        )
        trans_end = make_node(
            "Transpose",
            inputs=[f"{node.name}/Transpose/end_input0"],
            outputs=[node.output[0]],
            name=f"{node.name}/Transpose/end",
            perm=perm,
        )
        node.input[0] = trans_beg.output[0]
        node.output[0] = trans_end.input[0]
        self += [trans_beg, trans_end]
        self.set_attribute(node, "axis", -1)



"""
Copyright (C) 2024 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from contextlib import suppress
from typing import List, Literal, Optional, overload


def text_to_boolean(text: str | Literal["0", "1", "false", "true"]) -> bool:
    """Convert common text to bool

    Args:
        text (str): integer string or false/true literals
    """

    with suppress(ValueError):
        integer = int(text)
        return bool(integer)
    return text.lower() == "true"


@overload
def text_to_integers(text: None) -> None:
    """Convert comma-separated text to list of integers"""


@overload
def text_to_integers(text: str) -> List[int]:
    """Convert comma-separated text to list of integers"""


def text_to_integers(text: Optional[str]) -> List[int] | None:
    """Convert comma-separated text to list of integers

    Args:
        text (str | None): comma-separated integers
    """

    if text is None:
        return None
    elif text.strip() == "":
        return []
    return list(map(int, text.split(",")))



"""
Copyright (C) 2024 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import List

from onnx.onnx_pb import NodeProto, TensorProto

from . import IR_PASSES, OnnxGraph, Rewriter, SingleNodePattern


@IR_PASSES.register("eliminate_useless_floor")
class EliminateUselessFloorRewriter(Rewriter):
    """OpenVINO has a bug that floor on integers, it should be removed from the graph.

    Before:

        x(i64) -> Floor -> y(i64)

    After:

        x(i64) -> (Identity) -> y(i64)
    """

    def __init__(self):
        super().__init__(pattern=SingleNodePattern("Floor"))

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto]):
        node = nodes[0]
        _, input_type = graph.tensor_info(node.input[0])
        if input_type is None:
            return
        elif input_type in (
            TensorProto.FLOAT,
            TensorProto.FLOAT16,
            TensorProto.DOUBLE,
            TensorProto.BFLOAT16,
        ):
            return

        for next_n in self.get_output_node(node, 0):
            ind = 0
            for ind, input_name in enumerate(next_n.input):
                if input_name == node.output[0]:
                    break
            next_n.input[ind] = node.input[0]
        self -= node



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# pylint: disable=arguments-differ

from typing import List

import numpy as np
from onnx.helper import make_node
from onnx.onnx_pb import NodeProto

from ... import OnnxGraph
from .. import L2
from ..pattern import SingleNodePattern
from ..rewriter import Rewriter
from ..utils import make_constant


@L2.register("splittosequence_to_slice")
class SplitToSequenceToSliceRewriter(Rewriter):
    """Convert SplitToSequence and SequenceAt to slice operator."""

    def __init__(self):
        super().__init__(SingleNodePattern("SplitToSequence"))

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto]):
        node = nodes[0]
        sequence_consumers = graph.onnx_successors(node)
        axis = self.get_attribute(node, "axis", 0)
        keepdims = self.get_attribute(node, "keepdims", 1)
        assert isinstance(axis, int) and isinstance(keepdims, int)
        axis = int(axis)
        keepdims = bool(keepdims)
        input_shape = graph.tensor_shape(node.input[0])
        axis_dim = input_shape[axis]
        assert isinstance(axis_dim, int)

        if len(node.input) > 1:
            split_arr = self.get_value_or_die(node.input[1])
            if split_arr.ndim == 0:
                split = [int(split_arr)] * (axis_dim // int(split_arr))
            else:
                split = [int(i) for i in split_arr]
            assert sum(split) == input_shape[axis]
        else:
            split = [1] * axis_dim

        for i in sequence_consumers:
            if i.op_type != "SequenceAt":
                raise NotImplementedError(
                    f"{i.op_type} after SplitToSequence is not supported now"
                )

            self.sequence_at_to_slice(graph, i, split, axis, keepdims)
        # replace SplitToSequence with Identity
        self += make_node("Identity", [node.input[0]], [node.output[0]], name=node.name)
        self -= node

    def sequence_at_to_slice(
        self,
        graph: OnnxGraph,
        node: NodeProto,
        split: List[int],
        axis: int,
        keepdims: bool,
    ):
        """Convert SequenceAt to slice operator."""

        position = self.get_value_or_die(node.input[1])
        starts = sum(split[:position])
        ends = starts + split[position]

        starts_node = make_constant(
            name=f"{node.name}/Starts{position}", value=np.array([starts], "int64")
        )
        ends_node = make_constant(
            name=f"{node.name}/Ends{position}", value=np.array([ends], "int64")
        )
        axes_node = make_constant(
            name=f"{node.name}/Axes{position}", value=np.array([axis], "int64")
        )
        slice_node = make_node(
            "Slice",
            [
                node.input[0],
                starts_node.output[0],
                ends_node.output[0],
                axes_node.output[0],
            ],
            node.output,
            name=node.name,
        )
        if not keepdims and split[position] == 1:
            # add a squeeze
            squeeze_node = make_node(
                "Squeeze",
                [f"{node.name}/Squeeze_input0"],
                [node.output[0]],
                name=f"{node.name}/Squeeze",
            )
            slice_node.output[0] = squeeze_node.input[0]
            self += squeeze_node
        self += [starts_node, ends_node, axes_node, slice_node]
